
#include <assert.h>
#include <vector>
#include <utility>
#include <sstream>

#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_utility.hpp>

#include "Test.hpp"
#include "union_graph.hpp"

namespace boost
{
    // properties installation
    enum vertex_originality_t { vertex_originality = 101 };
    enum edge_ID_t { edge_ID = 102 };
    enum edge_originality_t { edge_originality = 103 };
    BOOST_INSTALL_PROPERTY(vertex, originality);
    BOOST_INSTALL_PROPERTY(edge, ID);
    BOOST_INSTALL_PROPERTY(edge, originality);
}

// properties definition
typedef boost::property<boost::vertex_name_t, std::string,
            boost::property<boost::vertex_originality_t, bool> > vertexs_properties;
typedef boost::property<boost::edge_ID_t, long,
            boost::property<boost::edge_weight_t, long,
                boost::property<boost::edge_originality_t, bool> > > edges_properties;

// graph definition
typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::directedS, vertexs_properties, edges_properties> Graph_vecS_vecS;
typedef boost::adjacency_list<boost::listS, boost::vecS, boost::directedS, vertexs_properties, edges_properties> Graph_listS_vecS;

typedef Graph_vecS_vecS Graph1;
typedef Graph_listS_vecS Graph2;

// templated edge vector; it contains vector_descriptor elements belong to the Graph template argument
template <typename Graph>
struct edge_vector {
    typedef std::vector<std::pair<typename boost::graph_traits<Graph>::vertex_descriptor,
                                  typename boost::graph_traits<Graph>::vertex_descriptor> > type;
    typedef typename type::iterator iterator;
    typedef typename type::const_iterator const_iterator;
};


class Test_union_graph : public Test
{
public:
	void startTest()
	{
        using namespace std;
        using namespace boost;

        // Setup Test
        const long NUM_VERTICES = 10;

        // creazione del grafo base
        Graph1 baseGraph( NUM_VERTICES );
        {
            graph_traits<Graph1>::edge_descriptor e;
            bool inserted;
            typedef property_map<Graph1, edge_weight_t>::type edge_weight_map_t;
            edge_weight_map_t edge_weight_map_base = get( edge_weight, baseGraph );
            for ( long i = 0; i < NUM_VERTICES ; ++i )
            {
                tie(e, inserted) = add_edge( i % NUM_VERTICES, (i + 1) % NUM_VERTICES, baseGraph );
                if ( inserted )
                    edge_weight_map_base[ e ] = 1;
            }
            cout << "Base graph:" << endl;
            print_graph( baseGraph );
            cout << endl;
        }

        // creazione del grafo piccolo
        Graph2 littleGraph( NUM_VERTICES );
        {
            graph_traits<Graph2>::edge_descriptor e;
            bool inserted;
            typedef property_map<Graph2, edge_weight_t>::type edge_weight_map_t;
            edge_weight_map_t edge_weight_map_little = get( edge_weight, littleGraph );
            edge_vector<Graph2>::type littleGraphEdgeVector;
            littleGraphEdgeVector.push_back( make_pair( 0, 6 ) );
            littleGraphEdgeVector.push_back( make_pair( 4, 9 ) );

            edge_vector<Graph2>::iterator vit;
            for ( vit = littleGraphEdgeVector.begin(); vit != littleGraphEdgeVector.end() ; ++vit )
            {
                tie(e, inserted) = add_edge( ( *vit ).first, ( *vit ).second, littleGraph );
                if ( inserted )
                    edge_weight_map_little[ e ] = 1;
            }
            cout << "Little graph:" << endl;
            print_graph( littleGraph );
            cout << endl;
        }

        // creazione dell'unione dei grafi
        typedef union_graph<Graph1, Graph2> UG;
        UG ug = make_union_graph( baseGraph, littleGraph );
        cout << "Union graph:" << endl;
        print_graph( ug );
        cout << endl;

        // vertice di test (0)
        graph_traits<UG>::vertex_descriptor u = make_WR1( graph_traits<Graph1>::vertex_descriptor( 0 ) );

        // Test 1: out_edges e out_degree
        {
            ostringstream oss(ostringstream::out);
            graph_traits<UG>::out_edge_iterator oei, oei_end;
            for ( tie( oei, oei_end ) = out_edges( u, ug ); oei != oei_end; ++oei )
                oss << target( *oei, ug );
            assert( "16" == oss.str() );
            assert( 2 == out_degree( u, ug ) );
        }
    /*
        // QUESTO TEST RICHIEDE LA BIDIREZIONALITA' (?)
        // SE E' VERO, ALLORA OCCORRE INSERIRE I CONTROLLI PER VEDERE SE CI SONO TUTTI I CONCEPTS
        // ALL'INTERNO DELLE FUNZIONI QUI SOTTO. USARE I CONTROLLI DELLE BOOST.

        // Test 2: in_edges, in_degree e degree
        {
            graph_traits<UG>::in_edge_iterator iei, iei_end;
            for ( tie( iei, iei_end ) = in_edges( u, ug ); iei != iei_end; ++iei )
            {
                graph_traits<UG>::vertex_descriptor s = source( *iei, ug );

                if ( boost::get<WR1<graph_traits<Graph1>::vertex_descriptor> >( &s ) )
                    cout << boost::get<WR1<graph_traits<Graph1>::vertex_descriptor> >( s )._data << endl;
                else
                    cout << boost::get<WR2<graph_traits<Graph2>::vertex_descriptor> >( s )._data << endl;
            }
            cout << "In degree: " << in_degree( u, ug ) << endl;
            cout << "Degree: " << degree( u, ug ) << endl;
        }
    */

        // Test 3: adjacent_vertices
        {
            ostringstream oss(ostringstream::out);
            graph_traits<UG>::adjacency_iterator ai, ai_end;
            for ( tie( ai, ai_end ) = adjacent_vertices( u, ug ); ai != ai_end; ++ai )
                oss << *ai;
            assert( "16" == oss.str() );
        }

        // Test 4: num_vertices, vertices
        {
            ostringstream oss(ostringstream::out);
            oss << num_vertices( ug );
            assert( "10" == oss.str() );
            oss.str("");

            graph_traits<UG>::vertex_iterator vi, vi_end;
            for ( tie( vi, vi_end ) = vertices( ug ); vi != vi_end; ++vi )
                oss << *vi;
            assert( "0123456789" == oss.str() );
        }

        // Test 5: num_edges, edges, source, target
        {
            ostringstream oss(ostringstream::out);
            oss << num_edges( ug );
            assert( "12" == oss.str() );
            oss.str("");

            graph_traits<UG>::edge_iterator ei, ei_end;
            for ( tie( ei, ei_end ) = edges( ug ); ei != ei_end; ++ei )
            {
                oss << source( *ei, ug ) << "->" << target( *ei, ug ) << "; ";
            }
            assert( "0->1; 1->2; 2->3; 3->4; 4->5; 5->6; 6->7; 7->8; 8->9; 9->0; 0->6; 4->9; " == oss.str() );
        }

        // Test 6: edge
        {
            {
                edge_vector<Graph1>::type existingEdgeVector;
                existingEdgeVector.push_back( make_pair( 0, 1 ) );
                existingEdgeVector.push_back( make_pair( 1, 2 ) );
                existingEdgeVector.push_back( make_pair( 2, 3 ) );
                existingEdgeVector.push_back( make_pair( 3, 4 ) );
                existingEdgeVector.push_back( make_pair( 4, 5 ) );
                existingEdgeVector.push_back( make_pair( 5, 6 ) );
                existingEdgeVector.push_back( make_pair( 6, 7 ) );
                existingEdgeVector.push_back( make_pair( 7, 8 ) );
                existingEdgeVector.push_back( make_pair( 8, 9 ) );
                existingEdgeVector.push_back( make_pair( 9, 0 ) );
                existingEdgeVector.push_back( make_pair( 0, 6 ) );
                existingEdgeVector.push_back( make_pair( 4, 9 ) );

                edge_vector<Graph1>::iterator vit;
                for ( vit = existingEdgeVector.begin(); vit != existingEdgeVector.end() ; ++vit )
                {
                    graph_traits<UG>::edge_descriptor e;
                    bool exists;
                    graph_traits<UG>::vertex_descriptor u1( make_WR1( graph_traits<Graph1>::vertex_descriptor( ( *vit ).first ) ) );
                    graph_traits<UG>::vertex_descriptor u2( make_WR1( graph_traits<Graph1>::vertex_descriptor( ( *vit ).second ) ) );
                    tie(e, exists) = edge( u1, u2, ug );
                    assert( exists );
                }
            }

            {
                edge_vector<Graph2>::type notExistingEdgeVector;
                notExistingEdgeVector.push_back( make_pair( 0, 2 ) );
                notExistingEdgeVector.push_back( make_pair( 5, 4 ) );
                notExistingEdgeVector.push_back( make_pair( 7, 2 ) );
                notExistingEdgeVector.push_back( make_pair( 4, 7 ) );
                notExistingEdgeVector.push_back( make_pair( 0, 4 ) );
                notExistingEdgeVector.push_back( make_pair( 5, 0 ) );
                notExistingEdgeVector.push_back( make_pair( 6, 2 ) );
                notExistingEdgeVector.push_back( make_pair( 4, 1 ) );
                notExistingEdgeVector.push_back( make_pair( 3, 8 ) );
                notExistingEdgeVector.push_back( make_pair( 9, 3 ) );
                notExistingEdgeVector.push_back( make_pair( 3, 6 ) );
                notExistingEdgeVector.push_back( make_pair( 9, 7 ) );

                edge_vector<Graph2>::iterator vit;
                for ( vit = notExistingEdgeVector.begin(); vit != notExistingEdgeVector.end() ; ++vit )
                {
                    graph_traits<UG>::edge_descriptor e;
                    bool exists;
                    graph_traits<UG>::vertex_descriptor u1( make_WR2( graph_traits<Graph2>::vertex_descriptor( ( *vit ).first ) ) );
                    graph_traits<UG>::vertex_descriptor u2( make_WR2( graph_traits<Graph2>::vertex_descriptor( ( *vit ).second ) ) );
                    tie(e, exists) = edge( u1, u2, ug );
                    assert( !exists );
                }
            }
        }

        // Test 7: vertex properties
        {
			typedef graph_traits<UG>::vertex_descriptor VDU;
			typedef graph_traits<Graph1>::vertex_descriptor VD1;
			typedef graph_traits<Graph2>::vertex_descriptor VD2;

			vector<VDU> uSet;	// [ VD1, VD2, VD1, VD2, ... ]
			for ( graph_traits<UG>::vertices_size_type i = 0; i < num_vertices( ug ); ++i )
			{
				if ( !( i % 2 ) )
					uSet.push_back( make_WR1( VD1( i ) ) );	// i even
				else
					uSet.push_back( make_WR2( VD2( i ) ) );	// i odd
			}

			// get( tag, g ), get( pmap, key )
            typedef property_map<UG, vertex_index_t>::type vertex_index_map_t;
            vertex_index_map_t indexmap = get( vertex_index, ug );
			for ( vector<VDU>::size_type i = 0; i < uSet.size(); ++i )
				assert( i == get( indexmap, uSet[ i ] ) );

			// get( tag, g ), put( pmap, key, value )
			// warning: don't "put" on vertex_index_t, it's read-only
            // NOTE: string( new char( 65 + char( i ) ), 1 ) :  string that contains the ASCII ( 65 + i )
			typedef property_map<UG, vertex_name_t>::type vertex_name_map_t;
            vertex_name_map_t namemap = get( vertex_name, ug );
			for ( vector<VDU>::size_type i = 0; i < uSet.size(); ++i )
				put( namemap, uSet[ i ], string( new char( 65 + char( i ) ), 1 ) );		// ok, ok... this is a leak...
			for ( vector<VDU>::size_type i = 0; i < uSet.size(); ++i )
				assert( string( new char( 65 + char( i ) ), 1 ) == get( namemap, uSet[ i ] ) );

			// get( tag, const g ), get( const pmap, key )
            typedef property_map<UG, vertex_index_t>::const_type vertex_index_map_const_t;
            vertex_index_map_const_t indexmap_const = get( vertex_index, const_cast<const UG&>( ug ) );
			for ( vector<VDU>::size_type i = 0; i < uSet.size(); ++i )
				assert( i == get( indexmap_const, uSet[ i ] ) );

			// get( tag, g, key )
			for ( vector<VDU>::size_type i = 0; i < uSet.size(); ++i )
				assert( i == get( vertex_index, ug, uSet[ i ] ) );

			// put( tag, g, key, value ) on vertex_name property
			for ( vector<VDU>::size_type i = 0; i < uSet.size(); ++i )
				put( vertex_name, ug, uSet[ i ], string( new char( 65 + char( i ) ), 1 ) );		// ok, ok... this is a leak...
			for ( vector<VDU>::size_type i = 0; i < uSet.size(); ++i )
				assert( string( new char( 65 + char( i ) ), 1 ) == get( vertex_name, ug, uSet[ i ] ) );

			// put( tag, g, key, value ) on vertex_originality property
			for ( vector<VDU>::size_type i = 0; i < uSet.size(); ++i )
				put( vertex_originality, ug, uSet[ i ], ( ( i%2 ) ? false : true ) );
			for ( vector<VDU>::size_type i = 0; i < uSet.size(); ++i )
				assert( ( ( i%2 ) ? false : true ) == get( vertex_originality, ug, uSet[ i ] ) );
        }

        // Test 8: edge properties
        {
			typedef graph_traits<UG>::edge_descriptor EDU;
			typedef graph_traits<Graph1>::edge_descriptor ED1;
			typedef graph_traits<Graph2>::edge_descriptor ED2;

			graph_traits<UG>::edge_iterator ei, ei_end;
			long i;

			// get( tag, g ), put( pmap, key, value ), get( pmap, key )
			typedef property_map<UG, edge_weight_t>::type edge_weight_map_t;
			edge_weight_map_t weightmap = get( edge_weight, ug );
			for ( tie( ei, ei_end) = edges( ug ), i = 0; ei != ei_end; ++ei )
				put( weightmap, *ei, i++ );
			for ( tie( ei, ei_end) = edges( ug ), i = 0; ei != ei_end; ++ei )
				assert( i++ == get( weightmap, *ei ) );

			// get( tag, const g ), get( const pmap, key )
            typedef property_map<UG, edge_weight_t>::const_type edge_weight_map_const_t;
            edge_weight_map_const_t weightmap_const = get( edge_weight, const_cast<const UG&>( ug ) );
            for ( tie( ei, ei_end) = edges( ug ), i = 0; ei != ei_end; ++ei )
                assert( i++ == get( weightmap_const, *ei ) );

			// get( tag, g, key )
            for ( tie( ei, ei_end) = edges( ug ), i = 0; ei != ei_end; ++ei )
                assert( i++ == get( edge_weight, ug, *ei ) );

			// put( tag, g, key, value ) on edge_weight property
            for ( tie( ei, ei_end) = edges( ug ), i = static_cast<long>( num_edges( ug ) ); ei != ei_end; ++ei )
				put( edge_weight, ug, *ei, i-- );
			for ( tie( ei, ei_end) = edges( ug ), i = static_cast<long>( num_edges( ug ) ); ei != ei_end; ++ei )
				assert( i-- == get( edge_weight, ug, *ei ) );

			// put( tag, g, key, value ) on edge_originality property
			for ( tie( ei, ei_end) = edges( ug ); ei != ei_end; ++ei )
				put( edge_originality, ug, *ei, ( ( i%2 ) ? false : true ) );
			for ( tie( ei, ei_end) = edges( ug ); ei != ei_end; ++ei )
				assert( ( ( i%2 ) ? false : true ) == get( edge_originality, ug, *ei ) );
        }

        // Test 9: EqualityComparable concept of edge_descriptor, requested by IncidenceGraph concept
        {
            typedef graph_traits<UG>::edge_descriptor EDU;
            typedef graph_traits<UG>::vertex_descriptor VDU;
			typedef graph_traits<Graph1>::edge_descriptor ED1;
			typedef graph_traits<Graph2>::edge_descriptor ED2;
			typedef graph_traits<Graph1>::vertex_descriptor VD1;
			typedef graph_traits<Graph2>::vertex_descriptor VD2;

            VDU v1 = make_WR1( VD1( 0 ) );
            VDU v2 = make_WR2( VD2( 1 ) );
            VDU v3 = make_WR1( VD1( 2 ) );

            bool existence;
            EDU e1, e2, e3;

            tie( e1, existence ) = edge( v1, v2, ug );
            tie( e2, existence ) = edge( v1, v2, ug );
            tie( e3, existence ) = edge( v2, v3, ug );

            assert( e1 == e2 );
            // assert( e2 != e3 );
        }

        cout << "All test passed" << endl;
    }
};