//=======================================================================
// Copyright 2009 Tellus s.r.l.
// Authors: Cosimo Calabrese
//
// Distributed under the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//=======================================================================

#ifndef BOOST_UNION_GRAPH__HPP
#define BOOST_UNION_GRAPH__HPP

#include <ostream>

#include <boost/graph/graph_traits.hpp>
#include <boost/graph/properties.hpp>
#include <boost/graph/adjacency_iterator.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/graph/exception.hpp>
#include <boost/variant.hpp>

#include "union_different_iterator.hpp"

namespace boost {

  //===========================================================================
  // wrappers

  template <typename T>
  class WR1
  {
  public:
    WR1() : _data( T() ) {}
    WR1( T data ) : _data( data ) {}
    inline const WR1<T> operator=( const WR1<T>& other )
    {
      _data = other._data;
      return(*this);
    }
    inline bool operator==( const WR1<T>& other ) const
    {
      return _data == other._data;
    }
    T _data;
  };

  template <typename T>
  inline WR1<T>
  make_WR1( T data ) {
    return WR1<T>( data );
  }

  template <typename T>
  class WR2
  {
  public:
    WR2() : _data( T() ) {}
    WR2( T data ) : _data( data ) {}
    inline const WR2<T> operator=( const WR2<T>& other )
    {
      _data = other._data;
      return(*this);
    }
    inline bool operator==( const WR2<T>& other ) const
    {
      return _data == other._data;
    }
    T _data;
  };

  template <typename T>
  inline WR2<T>
  make_WR2( T data ) {
    return WR2<T>( data );
  }

  //===========================================================================
  // exception for the preconditions

  struct not_same_directed_category : public bad_graph {
    not_same_directed_category()
      : bad_graph("The two graphs of an union_graph must have the same directed category.")
    { }
  };

  struct not_same_traversal_category : public bad_graph {
    not_same_traversal_category()
      : bad_graph("The two graphs of an union_graph must have the same traversal category.")
    { }
  };

  struct not_same_edge_parallel_category : public bad_graph {
    not_same_edge_parallel_category()
      : bad_graph("The two graphs of an union_graph must have the same edge parallel category.")
    { }
  };

  struct not_same_num_vertices : public bad_graph {
    not_same_num_vertices()
      : bad_graph("The two graphs of an union_graph must have the same number of vertices.")
    { }
  };

  //===========================================================================
  // union_graph

  struct union_graph_tag { };

  template <typename Graph1, typename Graph2>
  class union_graph {
    typedef union_graph<Graph1, Graph2> self;
  public:
    typedef Graph1 graph_type1;
    typedef Graph2 graph_type2;

    // Constructor
    union_graph( Graph1& g1, Graph2& g2 ) : m_g1(g1), m_g2(g2)
    {
        if ( !is_same<typename graph_traits<Graph1>::directed_category,
                      typename graph_traits<Graph2>::directed_category>::value )
          throw not_same_directed_category();

        if ( !is_same<typename graph_traits<Graph1>::traversal_category,
                      typename graph_traits<Graph2>::traversal_category>::value )
          throw not_same_traversal_category();

        if ( !is_same<typename graph_traits<Graph1>::edge_parallel_category,
                      typename graph_traits<Graph2>::edge_parallel_category>::value )
          throw not_same_edge_parallel_category();

        if ( num_vertices( g1 ) != num_vertices( g2 ) )
          throw not_same_num_vertices();
    }

    // Graph requirements
    typedef boost::variant<WR1<typename graph_traits<Graph1>::vertex_descriptor>,
                           WR2<typename graph_traits<Graph2>::vertex_descriptor> > vertex_descriptor;
    typedef typename graph_traits<Graph1>::directed_category          directed_category;
    typedef typename graph_traits<Graph1>::traversal_category         traversal_category;
    typedef typename graph_traits<Graph1>::edge_parallel_category     edge_parallel_category;

    // IncidenceGraph requirements
    typedef boost::variant<boost::WR1<typename boost::graph_traits<Graph1>::edge_descriptor>,
                           boost::WR2<typename boost::graph_traits<Graph2>::edge_descriptor> > edge_descriptor;
    typedef union_different_iterator< typename graph_traits<Graph1>::out_edge_iterator,
                                      typename graph_traits<Graph2>::out_edge_iterator,
                                      WR1, WR2 > out_edge_iterator;
    typedef typename graph_traits<Graph1>::degree_size_type           degree_size_type;

    // AdjacencyGraph requirements
    typedef union_different_iterator< typename graph_traits<Graph1>::adjacency_iterator,
                                      typename graph_traits<Graph2>::adjacency_iterator,
                                      WR1, WR2 > adjacency_iterator;

    // BidirectionalGraph requirements
    typedef union_different_iterator< typename graph_traits<Graph1>::in_edge_iterator,
                                      typename graph_traits<Graph2>::in_edge_iterator,
                                      WR1, WR2 > in_edge_iterator;

    // VertexListGraph requirements
    typedef union_different_iterator< typename graph_traits<Graph1>::vertex_iterator,
                                      typename graph_traits<Graph2>::vertex_iterator,
                                      WR1, WR2 > vertex_iterator;
    typedef typename graph_traits<Graph1>::vertices_size_type vertices_size_type;

    // EdgeListGraph requirements
    typedef union_different_iterator< typename graph_traits<Graph1>::edge_iterator,
                                      typename graph_traits<Graph2>::edge_iterator,
                                      WR1, WR2 > edge_iterator;

    typedef typename graph_traits<Graph1>::edges_size_type            edges_size_type;

    typedef typename ::boost::edge_property_type<Graph1>::type edge_property_type;
    typedef typename ::boost::vertex_property_type<Graph1>::type vertex_property_type;

    typedef union_graph_tag graph_tag;

#ifndef BOOST_GRAPH_NO_BUNDLED_PROPERTIES
    // Bundled properties support
    template<typename Descriptor>
    typename graph::detail::bundled_result<Graph1, Descriptor>::type&
    operator[](Descriptor x)
    { return const_cast<Graph1&>(this->m_g1)[x]; }

    template<typename Descriptor>
    typename graph::detail::bundled_result<Graph1, Descriptor>::type const&
    operator[](Descriptor x) const
    { return this->m_g1[x]; }
#endif // BOOST_GRAPH_NO_BUNDLED_PROPERTIES

    static vertex_descriptor null_vertex()
    {
       return Graph1::null_vertex();
    }

    // private:
    Graph1& m_g1;
    Graph2& m_g2;
  };

#ifndef BOOST_GRAPH_NO_BUNDLED_PROPERTIES
  template<typename Graph1, typename Graph2>
  struct vertex_bundle_type<union_graph<Graph1, Graph2> >
    : vertex_bundle_type<Graph1> { };

  template<typename Graph1, typename Graph2>
  struct edge_bundle_type<union_graph<Graph1, Graph2> >
    : edge_bundle_type<Graph1> { };
#endif // BOOST_GRAPH_NO_BUNDLED_PROPERTIES

  //===========================================================================
  // Non-member functions for the Union Graph

  // Helper functions
  template <typename Graph1, typename Graph2>
  inline union_graph<Graph1, Graph2>
  make_union_graph(Graph1& g1, Graph2& g2) {
    return union_graph<Graph1, Graph2>(g1, g2);
  }
/*
  template <typename Graph1, typename Graph2>
  inline union_graph<const Graph1, const Graph2>
  make_union_graph(const Graph1& g1, const Graph2& g2 ) {
    return union_graph<const Graph1, const Graph2>(g1, g2);
  }
*/
  namespace detail
  {
    template <typename Graph1, typename Graph2>
    std::pair<typename graph_traits<Graph1>::vertex_descriptor,
              typename graph_traits<Graph2>::vertex_descriptor>
    get_corresponding_vertices(const typename union_graph<Graph1, Graph2>::vertex_descriptor& u)
    {
      typedef union_graph<Graph1, Graph2> UG;
      typedef typename graph_traits<Graph1>::vertex_descriptor VD1;
      typedef typename graph_traits<Graph2>::vertex_descriptor VD2;

      VD1 u1;
      VD2 u2;
      if ( boost::get<WR1<VD1> >( &u ) )
      {
        u1 = boost::get<WR1<VD1> >( u )._data;
        u2 = u1;
      }
      else
      {
        u2 = boost::get<WR2<VD2> >( u )._data;
        u1 = u2;
      }
      return std::make_pair( u1, u2 );
    }
  }

  // IncidenceGraph concept requirements
  template <typename Graph1, typename Graph2>
  std::pair<typename union_graph<Graph1, Graph2>::out_edge_iterator,
            typename union_graph<Graph1, Graph2>::out_edge_iterator>
  out_edges(const typename union_graph<Graph1, Graph2>::vertex_descriptor& u,
            const union_graph<Graph1, Graph2>& g)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename graph_traits<Graph1>::vertex_descriptor VD1;
    typedef typename graph_traits<Graph2>::vertex_descriptor VD2;
    typedef typename graph_traits<Graph1>::out_edge_iterator OEI1;
    typedef typename graph_traits<Graph2>::out_edge_iterator OEI2;

    VD1 u1;
    VD2 u2;
    tie( u1, u2 ) = detail::get_corresponding_vertices<Graph1, Graph2>( u );
    return make_union_different_range<OEI1, OEI2, WR1, WR2>( out_edges(u1, g.m_g1), out_edges(u2, g.m_g2) );
  }

  template <typename Graph1, typename Graph2>
  typename union_graph<Graph1, Graph2>::vertex_descriptor
  source(const typename union_graph<Graph1, Graph2>::edge_descriptor& e,
         const union_graph<Graph1, Graph2>& g)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename graph_traits<Graph1>::vertex_descriptor VD1;
    typedef typename graph_traits<Graph2>::vertex_descriptor VD2;
    typedef typename graph_traits<Graph1>::edge_descriptor ED1;
    typedef typename graph_traits<Graph2>::edge_descriptor ED2;

    if ( boost::get<WR1<ED1> >( &e ) )
    {
      return typename graph_traits<UG>::vertex_descriptor( WR1<VD1>( source( boost::get<WR1<ED1> >( e )._data, g.m_g1 ) ) );
    }
    else
    {
      return typename graph_traits<UG>::vertex_descriptor( WR2<VD2>( source( boost::get<WR2<ED2> >( e )._data, g.m_g2 ) ) );
    }
  }

  template <typename Graph1, typename Graph2>
  typename union_graph<Graph1, Graph2>::vertex_descriptor
  target(const typename union_graph<Graph1, Graph2>::edge_descriptor& e,
         const union_graph<Graph1, Graph2>& g)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename graph_traits<Graph1>::vertex_descriptor VD1;
    typedef typename graph_traits<Graph2>::vertex_descriptor VD2;
    typedef typename graph_traits<Graph1>::edge_descriptor ED1;
    typedef typename graph_traits<Graph2>::edge_descriptor ED2;

    if ( boost::get<WR1<ED1> >( &e ) )
    {
      return typename graph_traits<UG>::vertex_descriptor( WR1<VD1>( target( boost::get<WR1<ED1> >( e )._data, g.m_g1 ) ) );
    }
    else
    {
      return typename graph_traits<UG>::vertex_descriptor( WR2<VD2>( target( boost::get<WR2<ED2> >( e )._data, g.m_g2 ) ) );
    }
  }

  template <typename Graph1, typename Graph2>
  typename union_graph<Graph1, Graph2>::degree_size_type
  out_degree(const typename union_graph<Graph1, Graph2>::vertex_descriptor& u,
             const union_graph<Graph1, Graph2>& g)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename graph_traits<Graph1>::vertex_descriptor VD1;
    typedef typename graph_traits<Graph2>::vertex_descriptor VD2;

    VD1 u1;
    VD2 u2;
    tie( u1, u2 ) = detail::get_corresponding_vertices<Graph1, Graph2>( u );
    return out_degree(u1, g.m_g1) + out_degree(u2, g.m_g2);
  }


  // BidirectionalGraph concept requirements
  template <typename Graph1, typename Graph2>
  std::pair<typename union_graph<Graph1, Graph2>::in_edge_iterator,
            typename union_graph<Graph1, Graph2>::in_edge_iterator>
  in_edges(const typename union_graph<Graph1, Graph2>::vertex_descriptor& u,
           const union_graph<Graph1, Graph2>& g)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename graph_traits<Graph1>::vertex_descriptor VD1;
    typedef typename graph_traits<Graph2>::vertex_descriptor VD2;
    typedef typename graph_traits<Graph1>::in_edge_iterator IEI1;
    typedef typename graph_traits<Graph2>::in_edge_iterator IEI2;

    VD1 u1;
    VD2 u2;
    tie( u1, u2 ) = detail::get_corresponding_vertices<Graph1, Graph2>( u );
    return make_union_different_range<IEI1, IEI2, WR1, WR2>( in_edges<Graph1>(u1, g.m_g1), in_edges<Graph2>(u2, g.m_g2) );
  }

  template <typename Graph1, typename Graph2>
  typename union_graph<Graph1, Graph2>::degree_size_type
  in_degree(const typename union_graph<Graph1, Graph2>::vertex_descriptor& u,
            const union_graph<Graph1, Graph2>& g)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename graph_traits<Graph1>::vertex_descriptor VD1;
    typedef typename graph_traits<Graph2>::vertex_descriptor VD2;

    VD1 u1;
    VD2 u2;
    tie( u1, u2 ) = detail::get_corresponding_vertices<Graph1, Graph2>( u );
    return in_degree(u1, g.m_g1) + in_degree(u2, g.m_g2);
  }

  template <typename Graph1, typename Graph2>
  inline typename union_graph<Graph1, Graph2>::degree_size_type
  degree(const typename union_graph<Graph1, Graph2>::vertex_descriptor& u,
         const union_graph<Graph1, Graph2>& g)
  {
    return out_degree(u, g) + in_degree(u, g);
  }

  // AdjacencyGraph concept requirements
  template <typename Graph1, typename Graph2>
  std::pair<typename union_graph<Graph1, Graph2>::adjacency_iterator,
            typename union_graph<Graph1, Graph2>::adjacency_iterator>
  adjacent_vertices(const typename union_graph<Graph1, Graph2>::vertex_descriptor& u,
                    const union_graph<Graph1, Graph2>& g)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename graph_traits<Graph1>::vertex_descriptor VD1;
    typedef typename graph_traits<Graph2>::vertex_descriptor VD2;
    typedef typename graph_traits<Graph1>::adjacency_iterator AI1;
    typedef typename graph_traits<Graph2>::adjacency_iterator AI2;

    VD1 u1;
    VD2 u2;
    tie( u1, u2 ) = detail::get_corresponding_vertices<Graph1, Graph2>( u );
    return make_union_different_range<AI1, AI2, WR1, WR2>( adjacent_vertices(u1, g.m_g1), adjacent_vertices(u2, g.m_g2) );
  }

  // VertexListGraph concept requirements
  template <typename Graph1, typename Graph2>
  inline typename union_graph<Graph1, Graph2>::vertices_size_type
  num_vertices(const union_graph<Graph1, Graph2>& g)
  {
    return num_vertices(g.m_g1);
  }

  template <typename Graph1, typename Graph2>
  std::pair<typename union_graph<Graph1, Graph2>::vertex_iterator,
            typename union_graph<Graph1, Graph2>::vertex_iterator>
  vertices(const union_graph<Graph1, Graph2>& g)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename graph_traits<Graph1>::vertex_iterator VI1;
    typedef typename graph_traits<Graph2>::vertex_iterator VI2;

    VI2 vi2 = VI2();
    VI2 vi2_end = vi2;
    return make_union_different_range<VI1, VI2, WR1, WR2>( vertices(g.m_g1), std::make_pair(vi2, vi2_end) );
  }

  // EdgeListGraph concept requirements
  template <typename Graph1, typename Graph2>
  inline typename union_graph<Graph1, Graph2>::edges_size_type
  num_edges(const union_graph<Graph1, Graph2>& g)
  {
    return num_edges(g.m_g1) + num_edges(g.m_g2);
  }

  template <typename Graph1, typename Graph2>
  inline std::pair<typename union_graph<Graph1, Graph2>::edge_iterator,
                   typename union_graph<Graph1, Graph2>::edge_iterator>
  edges(const union_graph<Graph1, Graph2>& g)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename graph_traits<Graph1>::edge_iterator EI1;
    typedef typename graph_traits<Graph2>::edge_iterator EI2;

    return make_union_different_range<EI1, EI2, WR1, WR2>( edges(g.m_g1), edges(g.m_g2) );
  }

  // AdjacencyMatrix concept requirements
  template <typename Graph1, typename Graph2>
  std::pair<typename union_graph<Graph1, Graph2>::edge_descriptor, bool>
  edge(const typename union_graph<Graph1, Graph2>::vertex_descriptor& u,
       const typename union_graph<Graph1, Graph2>::vertex_descriptor& v,
       const union_graph<Graph1, Graph2>& g)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename graph_traits<Graph1>::vertex_descriptor VD1;
    typedef typename graph_traits<Graph2>::vertex_descriptor VD2;
    typedef typename graph_traits<Graph1>::edge_descriptor ED1;
    typedef typename graph_traits<Graph2>::edge_descriptor ED2;

    VD1 u1, v1;
    VD2 u2, v2;
    tie( u1, u2 ) = detail::get_corresponding_vertices<Graph1, Graph2>( u );
    tie( v1, v2 ) = detail::get_corresponding_vertices<Graph1, Graph2>( v );

    // check existence on g1
    {
      ED1 e1;
      bool exists;
      tie( e1, exists ) = edge(u1, v1, g.m_g1);
      if ( exists )
        return std::make_pair( typename graph_traits<UG>::edge_descriptor( WR1<ED1>( e1 ) ), exists );
    }

    // check existence on g2
    {
      ED2 e2;
      bool exists;
      tie( e2, exists ) = edge(u2, v2, g.m_g2);
      return std::make_pair(
        typename graph_traits<UG>::edge_descriptor( WR2<ED2>( e2 ) ),
        exists);
    }
    // se l'edge esiste in g1, allora ritorno il pair estratto da g1;
    // altrimenti torno quello che estraggo da g2; non serve controllare
    // se esiste in g2, perch� se non esiste, edge() torna un
    // edge_descriptor vuoto + false
  }

  //===========================================================================
  // Property maps

  namespace detail {

    template <typename PMap1, typename PMap2, typename UnionGraph>
    struct union_vertex_property_map
      : public boost::put_get_helper<typename property_traits<PMap1>::reference,
                                     union_vertex_property_map<PMap1, PMap2, UnionGraph>
        >
    {
      typedef typename graph_traits<typename UnionGraph::graph_type1>::vertex_descriptor VD1;
      typedef typename graph_traits<typename UnionGraph::graph_type2>::vertex_descriptor VD2;
      typedef typename graph_traits<UnionGraph>::vertex_descriptor key_type;
      typedef typename property_traits<PMap1>::value_type value_type;
      typedef typename property_traits<PMap1>::reference reference;
      typedef boost::lvalue_property_map_tag category;

      inline union_vertex_property_map() { }
      inline union_vertex_property_map( const PMap1 pm1, const PMap2 pm2)
        : m_pm1( pm1 ), m_pm2( pm2 ) { }
      inline reference operator[]( const key_type& v ) const {
        if ( boost::get<WR1<VD1> >( &v ) )              // S: deve essere ricastato a "non const"? perch� il variant non contiene un "const key_type", ma un "keytype"
          return m_pm1[ ( get<WR1<VD1> >( v ) )._data ];
        else
          return m_pm2[ ( get<WR2<VD2> >( v ) )._data ];
      }
    private:
      const PMap1 m_pm1;
      const PMap2 m_pm2;
    };

    template <typename PMap1, typename PMap2, typename UnionGraph>
    struct union_edge_property_map
      : public boost::put_get_helper<typename property_traits<PMap1>::reference,
                                     union_edge_property_map<PMap1, PMap2, UnionGraph>
        >
    {
      typedef typename graph_traits<typename UnionGraph::graph_type1>::edge_descriptor ED1;
      typedef typename graph_traits<typename UnionGraph::graph_type2>::edge_descriptor ED2;
      typedef typename graph_traits<UnionGraph>::edge_descriptor key_type;
      typedef typename property_traits<PMap1>::value_type value_type;
      typedef typename property_traits<PMap1>::reference reference;
      typedef boost::lvalue_property_map_tag category;

      inline union_edge_property_map() { }
      inline union_edge_property_map( const PMap1 pm1, const PMap2 pm2)
        : m_pm1( pm1 ), m_pm2( pm2 ) { }
      inline reference operator[]( const key_type& e ) const {
        if ( boost::get<WR1<ED1> >( &e ) )
          return m_pm1[ get<WR1<ED1> >( e )._data ];
        else
          return m_pm2[ get<WR2<ED2> >( e )._data ];
      }
    private:
      const PMap1 m_pm1;
      const PMap2 m_pm2;
    };

    struct union_graph_vertex_property_selector {
      template <class UnionGraph, class Property, class PropertyTag>
      struct bind_ {
        typedef union_vertex_property_map<
                  typename property_map<typename UnionGraph::graph_type1, PropertyTag>::type,
                  typename property_map<typename UnionGraph::graph_type2, PropertyTag>::type,
                  UnionGraph
                > type;
        typedef union_vertex_property_map<
                  typename property_map<typename UnionGraph::graph_type1, PropertyTag>::const_type,
                  typename property_map<typename UnionGraph::graph_type2, PropertyTag>::const_type,
                  UnionGraph
                > const_type;
      };
    };

    struct union_graph_edge_property_selector {
      template <class UnionGraph, class Property, class PropertyTag>
      struct bind_ {
        typedef union_edge_property_map<
                  typename property_map<typename UnionGraph::graph_type1, PropertyTag>::type,
                  typename property_map<typename UnionGraph::graph_type2, PropertyTag>::type,
                  UnionGraph
                > type;
        typedef union_edge_property_map<
                  typename property_map<typename UnionGraph::graph_type1, PropertyTag>::const_type,
                  typename property_map<typename UnionGraph::graph_type2, PropertyTag>::const_type,
                  UnionGraph
                > const_type;
      };
    };
  } // namespace detail

  template <>
  struct vertex_property_selector<union_graph_tag> {
    typedef detail::union_graph_vertex_property_selector type;
  };
  template <>
  struct edge_property_selector<union_graph_tag> {
    typedef detail::union_graph_edge_property_selector type;
  };

  template <typename Graph1, typename Graph2, typename PropertyTag>
  inline typename property_map<union_graph<Graph1, Graph2>, PropertyTag>::type
  get(PropertyTag p, union_graph<Graph1, Graph2>& ug)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename property_map<UG, PropertyTag>::type PA;
    return PA( get( p, ug.m_g1 ), get( p, ug.m_g2 ) );
  }

  template <typename Graph1, typename Graph2, typename PropertyTag>
  inline typename property_map<union_graph<Graph1, Graph2>, PropertyTag>::const_type
  get(PropertyTag p, const union_graph<Graph1, Graph2>& ug)
  {
    typedef union_graph<Graph1, Graph2> UG;
    typedef typename property_map<UG, PropertyTag>::const_type PA;
    return PA( get( p, const_cast<const Graph1&>( ug.m_g1 ) ),      // S: forse devo prima castare a const il ug, e poi estrarre il grafo
               get( p, const_cast<const Graph2&>( ug.m_g2 ) ) );
  }

  template <typename Graph1, typename Graph2, typename Property, typename Key>
  inline typename property_map_value<union_graph<Graph1, Graph2>, Property>::type
  get(Property p, const union_graph<Graph1, Graph2>& ug, const Key& k)
  {
    typedef union_graph<Graph1, Graph2> UG;
    return get( get( p, const_cast<const UG&>( ug ) ), k );
  }

  template <typename Graph1, typename Graph2, typename Property,
            typename Key, typename Value>
  inline void
  put(Property p, const union_graph<Graph1, Graph2>& ug, const Key& k,
      const Value& val)
  {
    typedef union_graph<Graph1, Graph2> UG;
    put( get( p, const_cast<UG&>( ug ) ), k, val);
  }

  //===========================================================================
  // EqualityComparable concept, requested by IncidenceGraph concept

//   template <typename T1, typename T2>
//   bool operator!=( variant<WR1<T1>, WR2<T2> >& left, variant<WR1<T1>, WR2<T2> >& right )
//   {
//       return !( left == right );
//   }

    template <typename Graph1, typename Graph2>
    inline bool operator!=(
        const typename boost::graph_traits<union_graph<Graph1, Graph2> >::edge_descriptor& left,
        const typename boost::graph_traits<union_graph<Graph1, Graph2> >::edge_descriptor& right)
    { return !( left == right ); }
} // namespace boost


namespace std
{
  template <typename T1, typename T2>
  std::ostream& operator<<(std::ostream& os,
      const boost::variant<boost::WR1<T1>, boost::WR2<T2> >& v)
  {
    if ( boost::get<boost::WR1<T1> >( &v ) )
      os << boost::get<boost::WR1<T1> >( v )._data;
    else
      os << boost::get<boost::WR2<T2> >( v )._data;
    return os;
  }
}

#endif // BOOST_UNION_GRAPH__HPP
